#include <mpi.h>
#include <iostream>
#include <fstream>
#include <list>
#include <cstdlib>
#include <climits>

#include "tiles.h"

/* If this flag is defined, the master will do nothing but handling searches,
 * otherwise, the master will solve searches by itself, too. */
#define PURE_MASTER

using namespace std;

/* localized data which is used for the search by each node */
struct ProcessData
{
public:
	bool master;
	int rank;
	int size;
	Pattern P;
	int tileCount;
	int startDepth;

	/* used during a search */
	long long steps;
	int solutionCount;
	list<int*> *solutions;
	long curSearchId;

	/* these are just important for the master */
	long ignoreFlag;
	long endFlag;
	long serachInterval;

	MPI_Request *request;
	int recvValue;
	list<int*> *searches;
	long searchCount;
	long searchId;
	long long sumSteps;
	int sumSolutions;
	ofstream *out;
	int processesFinished;
};

/* tags for the MPI messages */
const int messageTag = 101;
const int requestTag = 102;
const int killTag = 103;

/* Called by master in order to send a search to one slave.
 * Sends the kill tag if there are no more searches left. */
inline void sendSearchToSlave(ProcessData* PD, int slaveRank)
{
	/* Check for remaining searches. */
	if (!PD->searches->empty())
	{
		int* p;
		/* Take next search from the list. */
		p = PD->searches->front();
		PD->searches->pop_front();
		PD->searchId++;

		/* output */
		cout << "Starting search " << PD->searchId << " on slave " 
			<< slaveRank << "." << endl;
		(*PD->out) << "Starting search " << PD->searchId << " on slave " 
			<< slaveRank << "." << endl;

		/* Send the start pattern of the search and the search id to the 
		 * slave. */
		MPI_Send(p, PD->startDepth, MPI_INT, slaveRank, messageTag, 
			MPI_COMM_WORLD);
		MPI_Send(&PD->searchId, 1, MPI_LONG, slaveRank, messageTag, 
			MPI_COMM_WORLD);

		delete[] p;
	}
	else
	{
		/* No more searches, send the kill tag to the slave. */
		MPI_Send(NULL, 0, MPI_INT, slaveRank, killTag, MPI_COMM_WORLD);
		PD->processesFinished++;
		
		/* output */
		cout << "Processes finished: " << PD->processesFinished << endl;
		(*PD->out) << "Processes finished: " << PD->processesFinished << endl;
	}
}

/* Called by master when one of the slaves finished a search. Since we listend
 * assynchronously, the value in recvValue now contains the number of solutions.
 * Outputs solutions if any were found. */
inline void recieveFromSlave(ProcessData *PD, int slaveRank)
{
	TileSet T;
	Pattern P;
	int *p;
	long long steps;
	MPI_Status status;
	long curSearchId;

	/* The first recieved value is the number of solutions. */
	PD->sumSolutions += PD->recvValue;

	/* Receive the search id from the client. */
	MPI_Recv(&curSearchId, 1, MPI_LONG, slaveRank, messageTag, MPI_COMM_WORLD,
		&status);

	/* Output which slave finished. */
	cout << "Slave " << slaveRank << " finished search " << curSearchId 
		<< ". Solutions found: " << PD->recvValue << " (overall " 
		<< PD->sumSolutions << ")" << endl;
	(*PD->out) << "Slave " << slaveRank << " finished search " << curSearchId 
		<< ". Solutions found: " << PD->recvValue << " (overall " 
		<< PD->sumSolutions << ")" << endl;

	/* Receive the number of steps in this search from the client. */
	MPI_Recv(&steps, 1, MPI_LONG_LONG, slaveRank, messageTag, MPI_COMM_WORLD,
		&status);
	PD->sumSteps += steps;

	/* Receive all the solutions that were found, if any. */
	for (int i = 0; i < PD->recvValue; ++i)
	{
		/* Receive one solution as tile assignment of the pattern. */
		p = new int[PD->P.getHeight()*PD->P.getWidth()];
		MPI_Recv(p, PD->P.getHeight()*PD->P.getWidth(), MPI_INT, slaveRank,
			messageTag, MPI_COMM_WORLD, &status);

		/* Copy tile assignment into pattern P. */
		P.copyFrom(&PD->P);
		P.copyAssignmentFrom(p);
		delete[] p;

		/* Generate the tile set T from the tile assignment. */
		T.generateFromAssignment(&P, PD->tileCount);

		/* Output solution. */
		cout << "==============================Solution"
			<< "==============================" << endl;
		cout << "Tiles:" << endl;
		cout << T.print() << endl;
		cout << "Assignment for pattern:" << endl;
		cout << P.printAs();
		cout << "=================================="
			<< "==================================" << endl;
		(*PD->out) << "==============================Solution"
			<< "==============================" << endl;
		(*PD->out) << "Tiles:" << endl;
		(*PD->out) << T.print() << endl;
		(*PD->out) << "Assignment for pattern:" << endl;
		(*PD->out) << P.printAs();
		(*PD->out)<< "=================================="
			<< "==================================" << endl;
	}

	/* Send another search to the slave. */
	sendSearchToSlave(PD, slaveRank);

	/* Setup master for asyncronous receiving of the next message. */
	MPI_Irecv(&PD->recvValue, 1, MPI_INT, MPI_ANY_SOURCE, requestTag, 
		MPI_COMM_WORLD, PD->request);
}


/* Recursively, places tile by tile into the assignment of pattern 'P'.
 * For each positions all tile types will be placed where the colors match and,
 * if the glues do not match, the tile will only be placed if its glues can be
 * altered to be matching, without destroying the property that 'T' is a DRTAS.
 */
void placeTile(ProcessData *PD, TileSet *T, int pos)
{
#ifndef PURE_MASTER
	/* The master checks for incomming messages once in a while.
	 * If pure master is set, then only the slaves run this part of the program
	 * and we do not have to check for incomming messages. */
	if (PD->master && (PD->steps % PD->serachInterval) == 0)
	{
		MPI_Status status;
		int flag = 0;

		MPI_Test(PD->request, &flag, &status);

		/* See if at least one slave needs a search. */
		while (flag)
		{
			/* Receive possible solutions from slave and then send it another
			 * search. */
			int slaveRank = status.MPI_SOURCE;
			recieveFromSlave(PD, slaveRank);

			/* Test if there are more slaves which finished there searches. */
			MPI_Test(PD->request, &flag, &status);
		}
	}
#endif

	PD->steps++;

	/* Get next position in which we have to place a tile. */
	int x = PD->P.getSearchX(pos);
	int y = PD->P.getSearchY(pos);

	if (x == -1)
	{
		/* -1 is only returned by 'getSearchX' if the last tile in the strategy
		 * has already been placed: a solution is found. */

		/* Save the solution as tile assignment. */
		int *p = new int[PD->P.getHeight()*PD->P.getWidth()];
		PD->P.copyAssignmentTo(p);
		PD->solutions->push_back(p);
		PD->solutionCount++;

		return;
	}

	/* Get input glues and color of the tile which we are about to place. */
	int inS, inW, color;
	color = PD->P.getPixel(x, y);

	if (y == 0 || PD->P.getPixel(x, y - 1) < 0)
		/* Does not have south neighbor. */
		inS = -1; 
	else
		inS = T->getTile(PD->P.getTileId(x, y - 1)).glue[north];

	if (x == 0 || PD->P.getPixel(x - 1, y) < 0)
		 /* Does not have west neighbor. */
		inW = -1;
	else
		inW = T->getTile(PD->P.getTileId(x - 1, y)).glue[east];

	/* Test whether or not there exists a tile type with these input glues.
	 * In this case, no other tile type can be placed. */
	int i = T->getIndexByInputGlues(inS, inW);

	if (i >= 0)
	{
		/* Tile with these input exists. */
		if (color == T->getTile(i).color)
		{
			/* Color matches, place tile and continue recursively. */
			PD->P.placeTile(x, y, i);
			placeTile(PD, T, pos + 1);
			PD->P.removeTile(x, y);
		}

		return;
	}

	/* Try every tile of the same color in this position, merge glues if 
	 * necessary. If a formerly unused tile is placed, we can stop the search 
	 * in order to prevent the program from generating a whole bunch of 
	 * isomorphic tile assemblies. */
	TileSet *S = new TileSet();
	bool used = true;

	for (i = 0; i < T->getSize() && used; ++i)
	{
		if (T->getTile(i).color < 0 || T->getTile(i).color == color)
		{
			/* Color matches: try to place the tile.
			 * The tile set will be altered in this step, so we make a copy. */
			S->copyFrom(T);
			
			/* Access to the current tile in tile set S. */
			Tile *t = &(S->getTile(i));

			if (!t->used)
			{
				/* This tile was formerly unused, thus, it has to be initialized
				 * with its new color. No more tiles will be placed afterwards
				 * as the 'used' flag is set to false.
				 * Note that the merging process below will always be successful
				 * in this case. */
				used = false;
				t->used = true;
				t->color = color;
			}

			/* Try to merge the new input glues of 't' with the old input glues
			 * of 't' throughout the tile set S. */
			if (S->mergeGluesV(inS, t->glue[south]) &&
				S->mergeGluesH(inW, t->glue[west]))
			{
				/* The tile can be placed and we continue recursively. */
				PD->P.placeTile(x, y, i);
				placeTile(PD, S, pos + 1);
				PD->P.removeTile(x, y);
			}
		}
	}

	delete S;
}

/* Only called by master in the initialization phase:
 * Basically does the same as 'placeTile', but after placing 'startDepth' tiles 
 * into the assembly, a new search is stored into a list. After finishing, this
 * list will contain all possible tile assignments for the patterns of the first
 * 'startDepth' positions. */
void placeTileInit(ProcessData* PD, TileSet *T, int pos)
{
	if (pos == PD->startDepth)
	{
		/* We save this assembly as start point for a later search and continue 
		 * the search. */
		PD->searchCount++;

		/* If a search range is given by ignoreFlag or endFlag, we don't save
		 * these assemblies. */
		if (PD->searchCount < PD->ignoreFlag)
		{
			PD->searchId++;
			return;
		}
		else if (PD->searchCount > PD->endFlag)
			return;

		/* Save the configuration as tile assignment. */
		int *p = new int[PD->startDepth];
		PD->P.copyAssignmentTo(p, PD->startDepth);
		PD->searches->push_back(p);

		return;
	}

	PD->sumSteps++;

	/* Get next position in which we have to place a tile. */
	int x = PD->P.getSearchX(pos);
	int y = PD->P.getSearchY(pos);

	/* Get input glues and color of the tile which we are about to place. */
	int inS, inW, color;
	color = PD->P.getPixel(x, y);

	if (y == 0 || PD->P.getPixel(x, y - 1) < 0)
		 /* Does not have south neighbor. */
		inS = -1;
	else
		inS = T->getTile(PD->P.getTileId(x, y - 1)).glue[north];

	if (x == 0 || PD->P.getPixel(x - 1, y) < 0)
		 /* Does not have west neighbor. */
		inW = -1;
	else
		inW = T->getTile(PD->P.getTileId(x - 1, y)).glue[east];

	/* Test whether or not there exists a tile type with these input glues.
	 * In this case, no other tile type can be placed. */
	int i = T->getIndexByInputGlues(inS, inW);

	if (i >= 0)
	{
		/* Tile with these input exists. */
		if (color == T->getTile(i).color)
		{
			/* Color matches, place tile and continue recursively. */
			PD->P.placeTile(x, y, i);
			placeTileInit(PD, T, pos + 1);
			PD->P.removeTile(x, y);
		}

		return;
	}

	/* Try every tile of the same color in this position, merge glues if 
	 * necessary. If an formerly unused tile is placed, we can stop the search
	 * in order to prevent the program from generating a whole bunch of 
	 * isomorphic tile assemblies. */
	TileSet *S = new TileSet();
	bool used = true;

	for (i = 0; i < T->getSize() && used; ++i)
	{
		if (T->getTile(i).color < 0 || T->getTile(i).color == color)
		{
			/* Color matches: try to place the tile.
			 * The tile set will be altered in this step, so we make a copy. */
			S->copyFrom(T);
			
			/* Access to the current tile in tile set S. */
			Tile *tile = &(S->getTile(i));

			if (!tile->used)
			{
				/* This tile was formerly unused, thus, it has to be initialized
				 * with its new color. No more tiles will be placed afterwards
				 * as the 'used' flag is set to false.
				 * Note that the merging process below will always be successful
				 * in this case. */
				used = false;
				tile->used = true;
				tile->color = color;
			}

			/* Try to merge the new input glues of 't' with the old input glues
			 * of 't' throughout the tile set. */
			if (S->mergeGluesV(inS, tile->glue[south]) &&
				S->mergeGluesH(inW, tile->glue[west]))
			{
				/* The tile can be placed and we continue recursively. */
				PD->P.placeTile(x, y, i);
				placeTileInit(PD, S, pos + 1);
				PD->P.removeTile(x, y);
			}
		}
	}

	delete S;
}


/* Run by the master. Sets up the list of all searches and sends out seraches to
 * the slave processes. */
void runMaster(ProcessData *PD)
{
	/* Init variables, which are used only by the master. */
	PD->searches = new list<int*>;
	PD->request = new MPI_Request;
	PD->searchCount = 0;
	PD->searchId = 0;
	PD->sumSteps = 0;
	PD->sumSolutions = 0;
	PD->processesFinished = 0;

	/* Output initial stats. */
	cout << "=================================="
		<< "==================================" << endl;
	cout << "Pattern:" << endl << PD->P.printPattern() << endl;
	cout << "Tile Count: " << PD->tileCount << endl;
	cout << "=================================="
		<< "==================================" << endl;
	(*PD->out) << "=================================="
		<< "==================================" << endl;
	(*PD->out) << "Pattern:" << endl << PD->P.printPattern() << endl;
	(*PD->out) << "Tile Count: " << PD->tileCount << endl;
	(*PD->out) << "=================================="
		<< "==================================" << endl;

	/* Initialize the list containing all searches. */
	TileSet *T = new TileSet();
	T->init(PD->tileCount);

	placeTileInit(PD, T, 0);

	/* Output the overall search number */
	cout << PD->searchCount << " searches initialized." << endl;
	(*PD->out) << PD->searchCount << " searches initialized." << endl;

	/* Send a search out to every slave. Also handles the killing of slaves,
	 * if there are more slaves than searches. */
	for (int i = 1; i < PD->size; ++i)
	{
		sendSearchToSlave(PD, i);
	}

	/* Setup MPI for assynchronous recieving for when the slaves are finished.
	 */
	MPI_Irecv(&PD->recvValue, 1, MPI_INT, MPI_ANY_SOURCE, requestTag, 
		MPI_COMM_WORLD, PD->request);

#ifndef PURE_MASTER
	/* If master works too, we have to get a search for it. */
	PD->solutions = new list<int*>;
	int *p;

	while (!PD->searches->empty())
	{
		/* Get one search from the list. */
		p = PD->searches->front();
		PD->searches->pop_front();
		
		/* Set up the tile assignment of the pattern. */
		PD->P.copyAssignmentFrom(p, PD->startDepth);
		delete[] p;
		
		/* Generate tile set from the tile assignment */
		T->generateFromAssignment(&PD->P, PD->tileCount);
		
		PD->steps = 0;
		PD->solutionCount = 0;

		/* Set search id and ouput. */
		PD->searchId++;
		PD->curSearchId = PD->searchId;
		cout << "Starting search " << PD->searchId << " on master." << endl;
		(*PD->out) << "Starting search " << PD->searchId << " on master." 
			<< endl;
		
		/* Start recursive search on master. */
		placeTile(PD, T, PD->startDepth);

		
		PD->sumSolutions += PD->solutionCount;
		PD->sumSteps += PD->steps;
		
		/* Output search result. */
		cout << "Master finished search " << PD->curSearchId 
			<< ". Solutions found: " << PD->solutionCount 
			<< " (overall " << PD->sumSolutions << ")" << endl;
		(*PD->out) << "Master finished search " << PD->curSearchId 
			<< ". Solutions found: " << PD->solutionCount
			<< " (overall " << PD->sumSolutions << ")" << endl;
			
		/* Output all the solutions that were found, if any. */
		while (!PD->solutions->empty())
		{
			/* Pop solution from list. */
			p = PD->solutions->front();
			PD->solutions->pop_front();
			
			/* Set up the tile assignment of the solution. */
			PD->P.copyAssignmentFrom(p);
			delete[] p;
			
			/* Generate the tile set. */
			T->generateFromAssignment(&PD->P, PD->tileCount);

			/* Output solution. */
			cout << "==============================Solution"
				<< "==============================" << endl;
			cout << "Tiles:" << endl;
			cout << T->print() << endl;
			cout << "Assignment for pattern:" << endl;
			cout << PD->P.printAs();
			cout << "=================================="
				<< "==================================" << endl;
			(*PD->out) << "==============================Solution"
				<< "==============================" << endl;
			(*PD->out) << "Tiles:" << endl;
			(*PD->out) << T->print() << endl;
			(*PD->out) << "Assignment for pattern:" << endl;
			(*PD->out) << PD->P.printAs();
			(*PD->out) << "=================================="
				<< "==================================" << endl;
		}
	}

	delete PD->solutions;
#else
	/* If the master is not working we need at least on slave. */
	assert(PD->size > 1);
#endif

	/* If the master is working this part is called after it finished its own
	 * searches. Otherwise, the master only waits for searches to finish and
	 * sends out more work. */
	int flag;
	MPI_Status status;

	/* Run until all slave processes are finished. */
	while (PD->processesFinished < PD->size - 1)
	{
		/* Wait until a message is received. */
		flag = 0;
		MPI_Test( PD->request, &flag, &status);

		if (flag)
		{
			/* One of the slave finished. Receive possible solutions and then
			 * send out a new search -- if searches are left. */
			int slaveRank = status.MPI_SOURCE;
			recieveFromSlave(PD, slaveRank);
		}
	}

	/* Output final summary. */
	cout << "=================================="
		<< "==================================" << endl;
	cout << "Overall solutions: " << PD->sumSolutions << endl;
	cout << "Assemblies generated: " << PD->sumSteps << endl;
	cout << "=================================="
		<< "==================================" << endl << endl;
	(*PD->out) << "=================================="
		<< "==================================" << endl;
	(*PD->out) << "Overall solutions: " << PD->sumSolutions << endl;
	(*PD->out) << "Assemblies generated: " << PD->sumSteps << endl;
	(*PD->out) << "=================================="
		<< "==================================" << endl << endl;

	/* There will still be one request for asynchronous receiving be left open.
	 * So we cancel that one. */
	MPI_Cancel(PD->request);

	/* Clean up. */
	delete T;
	delete PD->searches;
	delete PD->request;
}

/* Run on each of the slave nodes. Just wait for master to assign the work. */
void runSlaves(ProcessData *PD)
{
	/* Set up the variables for the slave node. */
	bool running = true;
	int *p = new int[PD->startDepth];
	int *s;
	TileSet *T = new TileSet();
	PD->solutions = new list<int*>;
	MPI_Status status;

	while (running)
	{
		/* Wait for the master to send a message. */
		MPI_Recv(p, PD->startDepth, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, 
			&status);

		/* Test whether the master send the kill tag or a search. */
		if (status.MPI_TAG != killTag)
		{
			/* Search was received. Get the searchId */
			MPI_Recv(&PD->curSearchId, 1, MPI_LONG, 0, messageTag, 
				MPI_COMM_WORLD, &status);

			/* Init the tile assignment. */
			PD->P.copyAssignmentFrom(p, PD->startDepth);
			
			/* Init the tile set. */
			T->generateFromAssignment(&PD->P, PD->tileCount);
			
			/* Init remaining variables. */
			PD->steps = 0;
			PD->solutionCount = 0;

			/* Start search. */
			placeTile(PD, T, PD->startDepth);

			/* Send statistics of the search back to master. */
			MPI_Send(&PD->solutionCount, 1, MPI_INT, 0, requestTag, 
				MPI_COMM_WORLD);
			MPI_Send(&PD->curSearchId, 1, MPI_LONG, 0, messageTag, 
				MPI_COMM_WORLD);
			MPI_Send(&PD->steps, 1, MPI_LONG_LONG, 0, messageTag, 
				MPI_COMM_WORLD);
			
			/* Send the solutions back to master, if any. */
			for (int i = 0; i < PD->solutionCount; ++i)
			{
				/* Get solution from list and send it back to master. */
				s = PD->solutions->front();
				PD->solutions->pop_front();
				MPI_Send(s, PD->P.getWidth()*PD->P.getHeight(), MPI_INT, 0, 
					messageTag, MPI_COMM_WORLD);
				delete[] s;
			}
			assert(PD->solutions->empty());
		}
		else
		{
			/* We received a kill tag, so we can finish. */
			running = false;
		}
	}

	/* Clean up. */
	delete T;
	delete[] p;
	delete PD->solutions;
}

/* Main Procedure */
int main(int argc, char **argv)
{
	/* Init MPI. */
	MPI_Init(&argc, &argv);

	ProcessData PD;
	MPI_Comm_rank(MPI_COMM_WORLD, &PD.rank);
	MPI_Comm_size(MPI_COMM_WORLD, &PD.size);

	/* Set tile count. */
	PD.tileCount = 13;

	/* number of tiles that are placed before the recursive search is started */
	PD.startDepth = 14;

	/* the pattern */
	string str;
	str = "";

	str += "110101111111011111111";
	str += "011110101101101111011";
	str += "000000000000000000000"; // blue
	str += "110101111111011111101";
	str += "011110101101101111011";
	str += "000000000000000000000"; // white
	str += "101011111110111110111";
	str += "011110101101101111011";
	str += "000000000000000000000"; // red
	str += "110111111011111101111";
	str += "011110101101101111011";
	str += "110111111011111101111";
	str += "011110101101101111011";
	str += "000000000000000000000"; // blue
	str += "110111111011111101111";
	str += "011110101101101111011";
	str += "110111111011111101111";
	str += "011110101101101111011";
	str += "000000000000000000000"; // blue
	str += "110111011011111101111";
	str += "011110101101101111011";
	str += "000000000000000000000"; // white
	str += "101101110111111011111";
	//	str += "01111010110110111011"; 
		// search is faster without last row -- probably

	PD.P.fromString(21, 23, str);

#if (0)
	/* Test pattern with only one solution. 
	 * Overwrites the previous setup. */
	PD.tileCount = 8;
	PD.startDepth = 6;
	str  = "01111100010011";
	str += "22222222222222";
	str += "01101111100010";
	str += "22222222222222";
	str += "01001101111100";
	str += "01001101111100";
	PD.P.fromString(14, 6, str);
#endif

	/* Initialize the search strategy. */
	PD.P.initStrategy();

	if (PD.rank == 0)
	{
		/* master */
		PD.master = true;

		/* The first argument can be used to skip a given number of search 
		 * threads if -- for example -- someone accidentally shut down their
		 * computer during a search. */
		if ( argc > 1 )
			PD.ignoreFlag = strtol(argv[1], NULL, 10);
		else
			PD.ignoreFlag = 0;	

		/* The second flag can be used to finish after a certain search 
		 * thread id. */
		if (argc > 2)
			PD.endFlag = strtol(argv[2], NULL, 10);
		else
			PD.endFlag = LONG_MAX;


		/* Amout of tiles that are placed before the master checks whether a
		 * slave is done: when having lots of slaves it is more important to 
		 * keep the slaves busy than being fast on the master node, in
		 * particular, when the startDepth is large. Note that this falg is
		 * pointless if we have a pure master. */
		PD.serachInterval = 1000;
		
		/* Log file: it's the same as the output on the standard stream. */
		PD.out = new ofstream("2Colors13Tiles.txt");
		assert(PD.out->is_open());

		/* start computation */
		runMaster( &PD );

		/* Close file. */
		PD.out->close();
		delete PD.out;
	}
	else
	{
		/* slaves */
		PD.master = false;

		runSlaves( &PD );
	}

	MPI_Finalize();
	return 0;
}
