/*
* tiles.h
*
*  Created on: 2013-08-26
*      Author: SK
*/

#ifndef TILES_H_
#define TILES_H_

#include <cstring>
#include <assert.h>
#include <sstream>


/* directions of the glues */
enum direction { north, east, south, west };

inline std::string printDirection(int D)
{
	switch (D)
	{
	case north: return "north";
		break;
	case east: return "east";
		break;
	case south: return "south";
		break;
	case west: return "west";
		break;
	}
	return "unknown";
}


/* colors of the tiles */
enum color { white, black, red, green, yellow, orange, purple, blue, cyan, gray };

inline std::string printColor(int c)
{
	switch (c)
	{
	case white: return "white";
		break;
	case black: return "black";
		break;
	case gray: return "gray";
		break;
	case red: return "red";
		break;
	case yellow: return "yellow";
		break;
	case orange: return "orange";
		break;
	case purple: return "purple";
		break;
	case blue: return "blue";
		break;
	case cyan: return "cyan";
		break;
	case green: return "green";
		break;
	}
	return "unknown";
}

/* This class stores a pattern as well as a tile assignment for the pattern.
 * It also generates a strategy in which order the tiles are placed in the 
 * assembly. */
class Pattern
{
protected:
	int *data;
	int *assignment;
	int *searchX;
	int *searchY;
	int h, w, n;

public:
	Pattern()
	{
		data = NULL;
		assignment = NULL;
		searchX = NULL;
		searchY = NULL;
		h = -1;
		w = -1;
		n = -1;
	}

	Pattern(int width, int height)
	{
		data = NULL;
		assignment = NULL;
		searchX = NULL;
		searchY = NULL;
		n = -1;
		init(width, height);
	}

	Pattern(Pattern *P)
	{
		data = NULL;
		assignment = NULL;
		searchX = NULL;
		searchY = NULL;

		copyFrom(P);
	}

	~Pattern()
	{
		if (data)
			delete[] data;
		if (assignment)
			delete[] assignment;
		if (searchX)
			delete[] searchX;
		if (searchY)
			delete[] searchY;
	}

	/* Creates a copy of the pattern P.
	 * P must have the pattern, assignment, and search strategy initialized,
	 * otherwise errors may occur.
	 */
	void copyFrom(Pattern *P)
	{
		/* Copy all fields. */
		h = P->h;
		w = P->w;
		n = P->n;

		assert(h > 0 && w > 0 && n > 0 && P->data && P->assignment);

		if (data)
			delete[] data;
		data = new int[w*h];
		memcpy(data, P->data, sizeof(int)*w*h);

		if (assignment)
			delete[] assignment;
		assignment = new int[w*h];
		memcpy(assignment, P->assignment, sizeof(int)*w*h);

		if (searchX)
			delete[] searchX;
		searchX = new int[n];
		memcpy(searchX, P->searchX, sizeof(int)*n);

		if (searchY)
			delete[] searchY;
		searchY = new int[n];
		memcpy(searchY, P->searchY, sizeof(int)*n);
	}

	/* Copies the tile assignment from P.
	 * P and this instance must have the same width and height, and the 
	 * assignment of this instance must be initialized; otherwise, functionality
	 * is not guaranteed. */
	void copyAssignmentFrom(Pattern *P)
	{
//		assert(w == P->w && h == P->h && assignment && P->assignment);
		memcpy(assignment, P->assignment, sizeof(int)*w*h);
	}

	/* Copies the first 'length' entries from the p into the assignment.
	 * All other fields are unassigned.
	 * If a negative 'length' is passed, then it is assumed that p contains w*h
	 * entries. */
	void copyAssignmentFrom(int *p, int length = -1)
	{
		if (length < 0)
			length = w*h;
//		else
//			assert(w*h >= length);

		/* Set every entry to -1. */
		for (int i = 0; i < w*h; ++i)
			assignment[i] = -1;

		/* Copy p to assignment */
		for (int i = 0; i < length; ++i)
			getTileId(getSearchX(i), getSearchY(i)) = p[i];
	}

	/* Copies the first 'length' entries from the assignment to 'p' according 
	 * to the search strategy.
	 * If a negative 'length' is passed, then all entries are copied. */
	void copyAssignmentTo(int *p, int length=-1)
	{
		if (length < 0)
			length = w*h;
//		else
//			assert(w*h >= length);

		for (int i = 0; i < length; ++i)
			p[i] = getTileId(getSearchX(i), getSearchY(i));
	}

	/* Generates an empty pattern and assignment (all values -1) of given size.
	 */
	void init(int width, int height)
	{
		w = width;
		h = height;

		if (data)
			delete[] data;
		data = new int[w*h];

		if (assignment)
			delete[] assignment;
		assignment = new int[w*h];

		/* Assign initial values. */
		for (int i = 0; i < w*h; ++i)
		{
			data[i] = -1;
			assignment[i] = -1;
		}
	}

	/* Initializes a tile placement strategy for the pattern.
	 * Start with position (0,0) and then successively place one partial column
	 * and one partial row such that a square is assembled. "Don't care" pixels
	 * (value -1) are ignored. When the square is as large as the width
	 * (resp. height) it grows row by row (resp. column by column).
	 *
	 * The first 16 tiles are placed like this:
	 * 9ABF
	 * 458E
	 * 137D
	 * 026C 
	 */
	void initStrategy()
	{
		if (searchX)
			delete[] searchX;
		if (searchY)
			delete[] searchY;

		/* Count all colored positions in the pattern (values != -1).
		 * This yields the amount of tiles that have to be placed. */
		n = 0;
		for (int i = 0; i < w*h; ++i)
			if (data[i] >= 0)
				n++;

		/* Init the search fields. */
		searchX = new int[n];
		searchY = new int[n];

		int pos = 0;

		int max;
		if (w > h)
			max = w;
		else
			max = h;

		/* Repeat for the maximum of width and height. */
		for (int i = 0; i < max; ++i)
		{
			/* Place one row and one column (if we did not reach a bound yet).
			 */
			if (i < h)
			{
				/* Add a partial row growing square. */
				for (int j = 0; j < i && j < w; ++j)
				{
					if (getPixel(j, i) >= 0)
					{
						searchX[pos] = j;
						searchY[pos] = i;
						++pos;
					}
				}
			}

			if (i < w)
			{
				/* Add a partial column growing square. */
				for (int j = 0; j <= i && j < h; ++j)
				{
					if (getPixel(i, j) >= 0)
					{
						searchX[pos] = i;
						searchY[pos] = j;
						++pos;
					}
				}
			}
		}
	}

	/* x-position of the i-th tile to place.
	 * Returns -1 if the strategy size is exceeded. */
	int getSearchX(int i)
	{
		if (i >= n)
			return -1;
		else
			return searchX[i];
	}

	/* y-position of the i-th tile to place.
 	* Returns -1 if the strategy size is exceeded. */
	int getSearchY(int i)
	{
		if (i >= n)
			return -1;
		else
			return searchY[i];
	}

	/* Generate the pattern from a string of size 'width'*'height'.
	 * The pattern has to be given row by row, starting with the top row,
	 * without any seperating characters.
	 * Values may be 0..9 or the first letter of color which is to be placed.
	 * x or X are seen as "Don't care" pixels, no tiles are placed in these
	 * positions. */
	void fromString(int width, int height, std::string pattern)
	{
		assert(pattern.length() == (unsigned)(width*height));

		init(width, height);

		char c;
		int i = 0;

		/* Read the string row by row.
		 * We start by reading the top row, position (0,0) will be lower left. 
		 */
		for (int y = h - 1; y >= 0; --y)
		{
			for (int x = 0; x < w; ++x)
			{
				c = pattern[x + y*w];

				if (c >= '0' && c <= '9')
					data[i] = c - '0';
				else if (c == 'w' || c == 'W')
					data[i] = white;
				else if (c == 'b' || c == 'B')
					data[i] = black;
				else if (c == 'g' || c == 'G')
					data[i] = green;
				else if (c == 'r' || c == 'R')
					data[i] = red;
				else if (c == 'y' || c == 'Y')
					data[i] = yellow;
				else if (c == 'o' || c == 'O')
					data[i] = orange;
				else if (c == 'p' || c == 'P')
					data[i] = purple;
				else if (c == 'c' || c == 'C')
					data[i] = cyan;
				else if (c == 'x' || c == 'X')
					data[i] = -1;
				else
					assert(false);

				++i;
			}
		}
	}

	/* Width of the pattern */
	int getWidth()
	{
		return w;
	}

	/* Height of the pattern */
	int getHeight()
	{
		return h;
	}

	/* Access position (x,y) in the pattern. */
	int& getPixel(int x, int y)
	{
		return data[x + y*w];
	}

	/* Access position (x,y) in the assignment. */
	int& getTileId(int x, int y)
	{
		return assignment[x + y*w];
	}

	/* Place a tile in the assignment at given positions. */
	void placeTile(int x, int y, int tile)
	{
		getTileId(x, y) = tile;
	}

	/* Remove a tile at the given position. */
	void removeTile(int x, int y)
	{
		getTileId(x, y) = -1;
	}

	/* Returns a string containing the pattern.
	 * The colors are represented by 0,...,9,A,...,Z. */
	std::string printPattern()
	{
		std::stringstream s;
		for (int y = h - 1; y >= 0; --y)
		{
			for (int x = 0; x < w; ++x)
			{
				if (getPixel(x, y) < 0)
				{
					s << "x";
				}
				else if (getPixel(x, y) >= 10)
				{
					s << (char)('A' - 10 + getPixel(x, y));
				}
				else
				{
					s << getPixel(x, y);
				}
			}
			s << std::endl;
		}
		return s.str();
	}

	/* Returns a string containing the assignment. */
	std::string printAs()
	{
		std::stringstream s;
		for (int y = h - 1; y >= 0; --y)
		{
			for (int x = 0; x < w; ++x)
			{
				if (getTileId(x, y) < 0)
				{
					s << "x";
				}
				else if (getTileId(x, y) >= 10)
				{
					s << (char)('A' - 10 + getTileId(x, y));
				}
				else
				{
					s << getTileId(x, y);
				}
			}
			s << std::endl;
		}
		return s.str();
	}
};


/* Stores one tile type. */
class Tile
{
public:
	int glue[4];
	int color;
	bool used;

	/* Returns a string containing the color of the tile type and all the glue 
	 * labels. Returns 'unused' if the tile is not used. */
	std::string print()
	{
		std::stringstream s;
		if (used)
		{
			s << "color: " << printColor(color);
			for (int i = 0; i < 4; ++i)
			{
				s << ", " << printDirection(i) << ": " << glue[i];
			}
		}
		else
			s << "unused";
		return s.str();
	}

	/* Returns true iff both tiles are not used or the color and all glues 
	 * are the same. */
	bool operator== (Tile &t)
	{
		return (!used && !t.used) ||
			(used && t.used && color == t.color &&
			glue[north] == t.glue[north] && glue[east] == t.glue[east] &&
			glue[south] == t.glue[south] && glue[west] == t.glue[west]);
	}

	bool operator!= (Tile &t)
	{
		return !(operator== (t));
	}
};


/* Contains one tile set. */
class TileSet
{
protected:
	Tile* tiles;
	int n;

public:
	TileSet()
	{
		tiles = NULL;
		n = -1;
	}

	~TileSet()
	{
		if (tiles)
			delete[] tiles;
	}

	/* Inits the tile set with 'count' tiles.
	 * All tiles are uncolored (color=-1), the used-flag is false,
	 * and the glues on all tiles are mututally distinct. */
	void init(int count)
	{
//		assert(count > 0);
		n = count;

		if (tiles)
			delete[] tiles;
		tiles = new Tile[count];

		/* Init the tiles with all different glues and without color. */
		for (int i = 0; i < count; ++i)
		{
			tiles[i].color = -1;
			tiles[i].used = false;
			for (int j = 0; j < 4; ++j)
				tiles[i].glue[j] = 4 * i + j;
		}
	}

	/* Returns the number of (used and unused) tiles in the set. */
	int getSize()
	{
		return n;
	}

	/* Access the i-th tile in the set. */
	Tile& operator [](int i)
	{
		return tiles[i];
	}

	/* Access the i-th tile in the set. */
	Tile& getTile(int i)
	{
		return tiles[i];
	}

	/* Returns the index of the tile with south input 's' and west input 'w'
	 * or -1 if such a tile does not exist. */
	int getIndexByInputGlues(int s, int w)
	{
		/* We use a simple linear search.
		 * This is faster than using maps or any other search structure we could
		 * come up with, for the small tile sets we are handling. */
		for (int i = 0; i < n; ++i)
			if (s == tiles[i].glue[south] && w == tiles[i].glue[west])
				return i;
		return -1;
	}

	/* Merges two east-west glue types to be the same.
	 * The output is true iff the merging did not result in an invalid tile set.
	 * If false is returned the tile set might be altered and is useless! */
	bool mergeGluesH(int a, int b)
	{
		/* Easy cases: negative values correspond to a glue on the seed or to a
		 * "don't care" tile. */
		if (a < 0 || b < 0 || a == b)
			return true;

		/* Always replace by the smaller value. */
		if (a > b)
			std::swap(a, b);

		/* Go through all tiles and replace east and west glues if they 
		 * match 'b'. */
		for (int i = 0; i < n; ++i)
		{
			if (tiles[i].glue[west] == b)
			{
				/* Test whether a tile with these inputs exists already.
				 * If so, merging would yield tile set which is not a DRTAS
				 * anymore. */
				if (getIndexByInputGlues(tiles[i].glue[south], a) >= 0)
					/* A tile with these input glues exists already. */
					return false;
				tiles[i].glue[west] = a;
			}

			/* East glues can always be replaced without effect on whether or
			 * not the tile set remains a DRTAS. */
			if (tiles[i].glue[east] == b)
				tiles[i].glue[east] = a;
		}
		return true;
	}

	/* Merges two north-south glue types to be the same.
	 * The output is true iff the merging did not result in an invalid tile set.
	 * If false is returned the tile set might be altered and is useless! */
	bool mergeGluesV(int a, int b)
	{
		/* Easy cases: negative values correspond to a glue on the seed or to a
		* "don't care" tile. */
		if (a < 0 || b < 0 || a == b)
			return true;

		/* Always replace by the smaller value. */
		if (a > b)
			std::swap(a, b);

		/* Go through all tiles and replace south and north glues if they 
		 * match 'b'. */
		for (int i = 0; i < n; ++i)
		{
			if (tiles[i].glue[south] == b)
			{
				/* Test whether a tile with these inputs exists already.
				 * If so, merging would yield tile set which is not a DRTAS 
				 * anymore. */
				if (getIndexByInputGlues(a, tiles[i].glue[west]) >= 0)
					/* A tile with these input glues exists already. */
					return false; 
				tiles[i].glue[south] = a;
			}

			/* North glues can always be replaced without effect on whether or
			 * not the tile set remains a DRTAS. */
			if (tiles[i].glue[north] == b)
				tiles[i].glue[north] = a;
		}
		return true;
	}


	/* Makes this instance a copy of the TileSet 'T'.
	 * It is required that this instance is either an empty tile set or
	 * that the size of tile set 'T' is as most the size of this instance. */
	void copyFrom(TileSet *T)
	{
//		assert(!tiles || n >= T->getSize());

		n = T->getSize();

		if (!tiles)
			tiles = new Tile[n];

		memcpy(tiles, T->tiles, sizeof(Tile)*n);
	}

	/* Returns a string containing all tiles in this set. */
	std::string print()
	{
		std::stringstream s;

		for (int i = 0; i < n; ++i)
		{
			s << "[";

			if (i >= 10) /* indices >= 10 are printed as capital letters */
				s << (char)('A' - 10 + i);
			else
				s << i;

			s << "]\t" << tiles[i].print() << std::endl;
		}

		return s.str();
	}

	/* Returns true iff both tile sets contain the same tile types. */
	bool operator== (TileSet &T)
	{
		if (T.getSize() != n)
			return false;

		for (int i = 0; i < n; ++i)
		if (T.getTile(i) != tiles[i])
			return false;
		return true;
	}

	/* Generates the tile set which is needed in order to tile the given
	 * assembly. Returns false if the tile set is not directed or color
	 * missmatches are found.
	 * If tileCount is negative, the number of tiles in the set is adjusted
	 * to the highest number in the assembly. */
	bool generateFromAssignment(Pattern *P, int tileCount = -1)
	{
		if (tileCount < 0)
		{
			tileCount = 0;
			/* Find largest tile index in assembly. */
			for (int x = 0; x < P->getWidth(); ++x)
			{
				for (int y = 0; y < P->getHeight(); ++y)
				{
					if (P->getTileId(x, y) > tileCount)
						tileCount = P->getTileId(x, y);
				}
			}
			tileCount++;
		}

		/* Initialize the standard tile set. */
		init(tileCount);

		int i, j, col;
		/* Get colors and set used tiles. */
		for (int x = 0; x < P->getWidth(); ++x)
		{
			for (int y = 0; y < P->getHeight(); ++y)
			{
				i = P->getTileId(x, y);
				col = P->getPixel(x, y);
				if (i >= 0)
				{
					/* Position is occupied by a tile. */
					if (!tiles[i].used)
					{
						tiles[i].color = col;
						tiles[i].used = true;
					}
					else if (tiles[i].color != col)
						/* The tile has been used in another position and the
						 * colors of the two positions do not match. */
						return false;
				}
			}
		}

		/* Merge all horizontal glues of neighbors. */
		for (int x = 0; x < P->getWidth() - 1; ++x)
		{
			for (int y = 0; y < P->getHeight(); ++y)
			{
				/* Get ideces of two neighbors. */
				i = P->getTileId(x, y);
				j = P->getTileId(x + 1, y);

				/* Only try the glue merging if both positions are occupied. */
				if (i >= 0 && j >= 0 &&
						!mergeGluesH(tiles[i].glue[east], tiles[j].glue[west]))
					/* Glues cannot be merged.
					 * The generated tile set is not a DRTAS. */
					return false;
			}
		}

		/* Merge all vertical glues of neighbors. */
		for (int x = 0; x < P->getWidth(); ++x)
		{
			for (int y = 0; y < P->getHeight() - 1; ++y)
			{
				/* Get ideces of two neighbors. */
				i = P->getTileId(x, y);
				j = P->getTileId(x, y + 1);

				/* Only try the glue merging if both positions are occupied. */
				if (i >= 0 && j >= 0 &&
						!mergeGluesV(tiles[i].glue[north], tiles[j].glue[south]))
					/* Glues cannot be merged.
						* The generated tile set is not a DRTAS. */
					return false;
			}
		}

		/* Everything went well. */
		return true;
	}
};



#endif /* TILES_H_ */
